package org.bdware.consistency.plugin.pbft;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.google.gson.JsonPrimitive;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.common.AbstractContextContractExecutor;
import org.bdware.sc.ComponedContractResult;
import org.bdware.sc.ContractResult;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.Node;
import org.bdware.sc.conn.OnHashCallback;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.consistency.Committer;
import org.bdware.sc.consistency.pbft.PBFTAlgorithm;
import org.bdware.sc.consistency.pbft.PBFTMember;
import org.bdware.sc.consistency.pbft.PBFTMessage;
import org.bdware.sc.consistency.pbft.PBFTType;
import org.bdware.sc.units.*;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.trustedmodel.MultiReqSeq;
import org.zz.gmhelper.SM2KeyPair;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

//TODO 追赶差下的调用
public class PBFTExecutor extends AbstractContextContractExecutor {
    private static final Logger LOGGER = LogManager.getLogger(PBFTExecutor.class);
    final Object lock = new Object();
    private final List<PubKeyNode> members;
    int resultCount;

    AtomicInteger request_index = new AtomicInteger(0);
    // key为requestID，value为其seq
    Map<String, MultiReqSeq> seqMap = new ConcurrentHashMap<>();
    Map<String, ResultCache> resultCache = new ConcurrentHashMap<>();
    //  MultiPointContractInfo info;
    String contractID;
    PBFTAlgorithm pbft;
    ContractCluster contractCluster;
    boolean isMaster;

    public PBFTExecutor(
            int c, String con_id, final String masterPubkey, String[] members) {
        resultCount = c;
        contractID = con_id;
        this.members = new ArrayList<>();
        isMaster = globalConf.getNodeID().equals(masterPubkey);
        pbft = new PBFTAlgorithm(isMaster);
        int count = 0;
        for (String mem : members) {
            PubKeyNode pubkeyNode = new PubKeyNode();
            pubkeyNode.pubkey = mem;
            PBFTMember pbftMember = new PBFTMember();
            pbftMember.isMaster = mem.equals(masterPubkey);
            pbft.addMember(pubkeyNode, pbftMember);
            this.members.add(pubkeyNode);
            if (globalConf.getNodeID().equals(mem)) {
                pbft.setSendID(count);
            }
            count++;
        }
        contractCluster = new ContractCluster(contractID, this.members);
        pbft.setConnection(contractCluster);
        final MultiContractMeta cei = cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID);
        pbft.setCommitter(new Committer() {
            @Override
            public void onCommit(ContractRequest data) {
                ResultCallback ret = null;
                final long startTime = System.currentTimeMillis();
                ret = new ResultCallback() {
                    @Override
                    public void onResult(String str) {
                        Map<String, String> ret = new HashMap<>();
                        ret.put("action", "receiveTrustfullyResult");
                        SM2KeyPair keyPair = globalConf.getKeyPair();
                        ret.put("nodeID", keyPair.getPublicKeyStr());
                        ret.put("responseID", data.getRequestID());
                        ret.put("executeTime", (System.currentTimeMillis() - startTime) + "");
                        ret.put("data", str);
                        cei.setLastExeSeq(data.seq);
                        networkManager.sendToAgent(masterPubkey, JsonUtil.toJson(ret));
                    }
                };
                cmActions.getManager().executeLocallyAsync(data, ret, null);
            }
        });
    }

    public void onSyncMessage(Node node, byte[] data) {

        pbft.onMessage(node, data);
    }

    public void setSeq(int seq) {
        request_index = new AtomicInteger(seq);
        pbft.setAtomSeq(request_index.get());
    }

    public ResultCallback createResultCallback(
            final String requestID,
            final ResultCallback originalCb,
            final int count,
            final int request_seq,
            final String contractID) {
        ComponedContractResult componedContractResult = new ComponedContractResult(count);
        // TODO 加对应的超时？
        return networkManager.createResultCallback(
                requestID, new ResultMerger(originalCb, count, request_seq, contractID), count);
    }

    public void sendRequest(String id, ContractRequest req, ResultCallback collector) {
//        Map<String, Object> reqStr = new HashMap<>();
//        reqStr.put("uniReqID", id);
//        reqStr.put("data", req);
//        reqStr.put("action", "executeContractLocally");
        ContractRequest cr2 = ContractRequest.parse(req.toByte());
        cr2.setRequestID(id);
        PBFTMessage request = new PBFTMessage();
        request.setOrder(req.seq);
        request.setType(PBFTType.Request);
        request.setContent(cr2.toByte());
        for (PubKeyNode node : members) {
            if (!networkManager.hasAgentConnection(node.pubkey)) {
                LOGGER.warn("cmNode " + node.pubkey.substring(0, 5) + " is null");
                collector.onResult(
                        "{\"status\":\"Error\",\"result\":\"node offline\","
                                + "\"nodeID\":\""
                                + node
                                + "\","
                                + "\"action\":\"onExecuteContractTrustfully\"}");
//            } else if (MasterServerRecoverMechAction.recoverStatus.get(node).get(contractID)
//                    != RecoverFlag.Fine) {
//                collector.onResult(
//                        "{\"status\":\"Error\",\"result\":\"node recovering\","
//                                + "\"nodeID\":\""
//                                + node
//                                + "\","
//                                + "\"action\":\"onExecuteContractTrustfully\"}");
//                contractCluster.sendMessage(node, request.getBytes());
            } else {
                contractCluster.sendMessage(node, request.getBytes());
            }
        }
        // master负责缓存请求
        if (!masterServerTCPAction.getReqCache().containsKey(contractID)) {
            masterServerTCPAction.getReqCache().put(contractID, new RequestCache());
        }
        // TODO 多调多统一个seq的有多个请求，这个需要改
        String[] nodes =
                cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID).getMembers();
        LOGGER.info("cluster size=" + nodes.length + " contract " + req.getContractID());
        LOGGER.debug("contract " + req.getContractID() + " cluster: " + JsonUtil.toJson(nodes));

    }


    public boolean checkCurNodeNumValid() {
        return true;
    }

    @Override
    public void execute(String requestID, ContractRequest req, ResultCallback rc, OnHashCallback hcb) {
        LOGGER.debug(JsonUtil.toJson(req));
        MultiContractMeta meta = cmActions.getManager().multiContractRecorder.getMultiContractMeta(req.getContractID());
        if (meta == null || !meta.isMaster()) {
            cmActions.getManager().executeContractOnOtherNodes(req, rc);
            return;
        }
        req.setContractID(cmActions.getManager().getContractIDByName(req.getContractID()));
        // 三个相同requestID进来的时候，会有冲突。
        // 仅在此处有冲突么？
        // 这里是从MasterServer->MasterClient，请求的是"executeContractLocally"。

        // 如果是多点合约的请求，A1、A2、A3的序号应该一致，不能分配一个新的seq，根据requestID判断是否不需要重新分配一个序号
        //TODO seqMap memory leak
        //TODO
        //TODO
        if (null != requestID && requestID.endsWith("_mul")) {
            synchronized (lock) {
                if (seqMap.containsKey(requestID)) {
                    req.seq = seqMap.get(requestID).seq;
                } else {
                    req.seq = request_index.getAndIncrement();
                    seqMap.put(requestID, new MultiReqSeq(req.seq));
                }
            }
        } else {
            req.seq = request_index.getAndIncrement();
        }
        req.needSeq = true;
        String id =
                System.currentTimeMillis() + "_" + (int) (Math.random() * 1000000) + "_" + req.seq;
        LOGGER.info("execute receive requestID=" + requestID + "  msgID=" + id);

        if (checkCurNodeNumValid()) {
            LOGGER.debug("checkCurNodeNumValid=true");
            ResultCallback collector =
                    createResultCallback(id, rc, resultCount, req.seq, req.getContractID());
            masterServerTCPAction.getSync().sleep(id, collector);
            LOGGER.info("requestID=" + requestID + "  master broadcasts request " + req.seq);
            sendRequest(id, req, collector);
        } else {
            LOGGER.debug("invalidNodeNumOnResult");
            request_index.getAndDecrement();
            ContractResult finalResult =
                    new ContractResult(
                            ContractResult.Status.Error,
                            new JsonPrimitive("node number unavailable, request refused."));
            rc.onResult(JsonUtil.toJson(finalResult));
        }

        // }

        /*        // 三个相同requestID进来的时候，会有冲突。
        // 仅在此处有冲突么？
        // 这里是从MasterServer->MasterClient，请求的是"executeContractLocally"。
        req.seq = request_index.getAndIncrement();
        req.needSeq = true;
        ResultCallback collector = createResultCallback(id, rc, resultCount, req.getContractID());
        MasterServerTCPAction.sync.sleep(id, collector);
        sendRequest(id, req, collector);*/
    }

    // 清理缓存的多点合约请求序号
    public void clearCache() {
        final long time = System.currentTimeMillis() - 30000L;
        seqMap.entrySet()
                .removeIf(
                        entry -> {
                            MultiReqSeq cache = entry.getValue();
                            if (null == cache) {
                                return true;
                            }
                            return cache.startTime < time;
                        });
    }

    public static class ResultMerger extends ResultCallback {
        ComponedContractResult componedContractResult;
        AtomicInteger order;
        String contractID;
        int count;
        int request_seq;
        ResultCallback originalCallback;
        Set<String> nodeIDs = new HashSet<>(); // 已收到返回结果的节点

        ResultMerger(
                final ResultCallback originalCb,
                final int count,
                final int request_seq,
                final String contractID) {
            originalCallback = originalCb;
            this.count = count;
            this.request_seq = request_seq;
            this.contractID = contractID;
            componedContractResult = new ComponedContractResult(count);
            order = new AtomicInteger(0);
        }

        public String getContractID() {
            return contractID;
        }

        public String getInfo() {
            return "contractID="
                    + contractID
                    + "  收到第 "
                    + order
                    + " 个节点回复 : "
                    + "  order="
                    + order
                    + "   count="
                    + count
                    + "  ";
        }

        @Override
        public void onResult(String str) {
            // TODO 必须在这里聚合。
            // str的data是个ContractResult
            // 在这儿也是返回个ContractResult
            try {
                LOGGER.debug("a result of contract" + contractID + ": " + str);
                JsonObject obj = JsonParser.parseString(str).getAsJsonObject();
                if (obj.has("nodeID")) {
                    String id = obj.get("nodeID").getAsString();
                    if (nodeIDs.contains(id)) {
                        LOGGER.debug(
                                "ignored result because the result of node "
                                        + id.substring(0, 5)
                                        + " has been received");
                        return;
                    }
                    nodeIDs.add(id);
                }

                LOGGER.debug(
                        String.format(
                                "contractID=%s  received=%s  order=%d  count=%d",
                                contractID, str, order.get(), count));
                componedContractResult.add(obj);
                // 收集到所有结果
                if (order.incrementAndGet() == count) {
                    ContractResult finalResult = componedContractResult.figureFinalResult();
                    finalResult.needSeq = true;
                    finalResult.seq = request_seq;

                    //                    if (null == finalResult) {
                    //                        finalResult =
                    //                                new ContractResult(
                    //                                        ContractResult.Status.Exception,
                    //                                        new JsonPrimitive(
                    //                                                "no nore than half of the
                    // consistent result"));
                    //                        originalCallback.onResult(new
                    // Gson().toJson(finalResult));
                    //                    } else {
                    originalCallback.onResult(JsonUtil.toJson(finalResult));
                    //                    }
                    LOGGER.debug(
                            String.format(
                                    "%d results are the same: %s",
                                    finalResult.size, finalResult.result));

                    // 集群中事务序号+1
                    cmActions.getManager().multiContractRecorder
                            .getMultiContractMeta(contractID)
                            .nextSeqAtMaster();

                    // recover，其中无状态合约CP出错无需恢复
                    Set<String> nodesID = componedContractResult.getProblemNodes();
                    if (null == nodesID || nodesID.isEmpty()) {
                        return;
                    }
                    for (String nodeID : nodesID) {
                        LOGGER.warn("node fails! " + nodeID);
                        if (masterServerRecoverMechAction.getRecoverStatusMap().get(nodeID).get(contractID)
                                == RecoverFlag.Fine) {
                            masterServerRecoverMechAction.getRecoverStatusMap()
                                    .get(nodeID)
                                    .put(contractID, RecoverFlag.ToRecover);
                        }
                    }
                    for (String nodeID : nodesID) {
                        if (masterServerRecoverMechAction.getRecoverStatusMap().get(nodeID).get(contractID)
                                == RecoverFlag.ToRecover) {
                            LOGGER.warn("node in recover " + nodeID);

                            // 因为该节点结果有误，所以即时是stableMode也认为trans记录不可信
                            // 直接通过load别的节点来恢复
                            masterServerRecoverMechAction.restartContractFromCommonMode(
                                    nodeID, contractID);
                        }
                    }
                }
                //  clearCache();
            } catch (Exception e) {
                e.printStackTrace();
                LOGGER.warn("result exception!");
            }
        }
    }

    @Override
    public void onRecover(Map<String, Object> args) {
        int ceiLastExeSeq = (int) args.get("ceiLastExeSeq");
        this.setSeq(ceiLastExeSeq + 1);
    }
}
