package org.bdware.consistency.plugin.raft;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.common.AbstractContextContractExecutor;
import org.bdware.consistency.plugin.pbft.ContractCluster;
import org.bdware.consistency.plugin.ra.RequestAllExecutor;
import org.bdware.consistency.plugin.raft.algo.RaftAlgorithm;
import org.bdware.consistency.plugin.raft.algo.RaftConfig;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.Node;
import org.bdware.sc.conn.OnHashCallback;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.consistency.Committer;
import org.bdware.sc.units.MultiContractMeta;
import org.bdware.sc.units.PubKeyNode;
import org.bdware.sc.util.JsonUtil;
import org.zz.gmhelper.SM2KeyPair;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class RAFTExecutor extends AbstractContextContractExecutor {
    private static final Logger LOGGER = LogManager.getLogger(RAFTExecutor.class);
    private final ContractCluster contractCluster;
    RaftAlgorithm raft;

    public RAFTExecutor(String contractID, String[] memberPubKeys) {

        RaftConfig config = new RaftConfig();
        PubKeyNode selfNode = new PubKeyNode();
        selfNode.pubkey = globalConf.getNodeID();
        List<PubKeyNode> members = new ArrayList<>();
        for (String memberPubKey : memberPubKeys) {
            PubKeyNode pubkeyNode = new PubKeyNode();
            pubkeyNode.pubkey = memberPubKey;
            members.add(pubkeyNode);
        }
        contractCluster = new ContractCluster(contractID, members);
        final MultiContractMeta cei = cmActions.getManager().multiContractRecorder.getMultiContractMeta(contractID);
        raft = new RaftAlgorithm(config, selfNode, contractCluster, new Committer() {
            @Override
            public void onCommit(ContractRequest request) {
                ResultCallback ret = null;
                final long startTime = System.currentTimeMillis();
                ret = new ResultCallback() {
                    @Override
                    public void onResult(String str) {
                        Map<String, String> ret = new HashMap<>();
                        ret.put("action", "receiveTrustfullyResult");
                        SM2KeyPair keyPair = globalConf.getKeyPair();
                        ret.put("nodeID", keyPair.getPublicKeyStr());
                        ret.put("responseID", request.getRequestID());
                        ret.put("executeTime", (System.currentTimeMillis() - startTime) + "");
                        ret.put("data", str);
                        cei.setLastExeSeq(request.seq);
                        networkManager.sendToAgent(raft.getLeader().pubkey, JsonUtil.toJson(ret));
                    }
                };
                cmActions.getManager().executeLocallyAsync(request, ret, null);
            }
        });

        //TODO 在改变角色时需要通知node cluster
        if (cei.isMaster()) {
            raft.convertToLeader();
        } else {
            raft.convertToFollower();
        }
    }

    @Override
    public void execute(String requestID, ContractRequest req, ResultCallback rcb, OnHashCallback hcb) {
        LOGGER.debug(JsonUtil.toJson(req));
        MultiContractMeta multiContractMeta = cmActions.getManager().multiContractRecorder.getMultiContractMeta(req.getContractID());
        if (multiContractMeta == null || !multiContractMeta.isMaster()) {
            cmActions.getManager().executeContractOnOtherNodes(req, rcb);
            return;
        }

        ResultCallback collector;
        int count = contractCluster.getNodes().size();
        collector = networkManager.createResultCallback(
                requestID, new RequestAllExecutor.ResultMerger(rcb, count, req.seq, req.getContractID()), count);
        masterServerTCPAction.getSync().sleep(requestID, collector);
        raft.insertLogEntry(req);
    }


    @Override
    public void onRecover(Map<String, Object> args) {
        super.onRecover(args);
    }

    @Override
    public void onDeliverBlock(String data) {
        super.onDeliverBlock(data);
    }

    @Override
    public void close() {
        super.close();
    }

    @Override
    public void onSyncMessage(Node node, byte[] data) {
        raft.onMessage(node, data);
    }
}
