package org.bdware.consistency.plugin.raft.algo.message;

import org.bdware.sc.conn.ByteUtil;
import org.bdware.sc.units.PubKeyNode;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

public class RequestVote implements IRaftMessage {
    RaftMessageType type;
    long term;
    PubKeyNode candidate; // candidate's Id
    long lastLogIndex;
    long lastLogTerm;

    public RequestVote() {
        type = RaftMessageType.RequestVote;
    }

    public RequestVote(long term, PubKeyNode candidate, long lastLogIndex, long lastLogTerm) {
        type = RaftMessageType.RequestVote;
        this.term = term;
        this.candidate = candidate;
        this.lastLogIndex = lastLogIndex;
        this.lastLogTerm = lastLogTerm;
    }

    @Override
    public byte[] getBytes() {
        ByteArrayOutputStream msg = new ByteArrayOutputStream();
        try {
            ByteUtil.writeLong(msg, term);
            if (candidate != null && !candidate.pubkey.equals("")) {
                byte[] pubkey = candidate.pubkey.getBytes();
                ByteUtil.writeInt(msg, pubkey.length);
                msg.write(pubkey);
            } else {
                ByteUtil.writeInt(msg, 0);
            }
            ByteUtil.writeLong(msg, lastLogIndex);
            ByteUtil.writeLong(msg, lastLogTerm);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return msg.toByteArray();
    }

    @Override
    public IRaftMessage parse(byte[] b) {
        RequestVote msg = new RequestVote();
        ByteArrayInputStream bi = new ByteArrayInputStream(b);
        msg.term = ByteUtil.readLong(bi);
        int candidateLen = ByteUtil.readInt(bi);
        PubKeyNode node = new PubKeyNode();
        if (candidateLen > 0) {
            node.pubkey = new String(ByteUtil.readBytes(bi, candidateLen));
        }
        msg.candidate = node;
        msg.lastLogIndex = ByteUtil.readLong(bi);
        msg.lastLogTerm = ByteUtil.readLong(bi);
        return msg;
    }

    @Override
    public RaftMessageType getType() {
        return type;
    }

    public long getTerm() {
        return term;
    }

    public void setTerm(long term) {
        this.term = term;
    }

    public PubKeyNode getCandidate() {
        return candidate;
    }

    public void setCandidate(PubKeyNode candidate) {
        this.candidate = candidate;
    }

    public long getLastLogIndex() {
        return lastLogIndex;
    }

    public void setLastLogIndex(long lastLogIndex) {
        this.lastLogIndex = lastLogIndex;
    }

    public long getLastLogTerm() {
        return lastLogTerm;
    }

    public void setLastLogTerm(long lastLogTerm) {
        this.lastLogTerm = lastLogTerm;
    }
}
