package org.bdware.consistency.plugin.raft.algo.message;

import org.bdware.consistency.plugin.raft.algo.storage.LogEntry;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.units.PubKeyNode;
import org.junit.Test;

import java.util.ArrayList;
import java.util.List;

public class MessageTest {
    @Test
    public void testRequestVote() {
        RequestVote msg = new RequestVote();
        msg.setTerm(98567);
        PubKeyNode node = new PubKeyNode();
        node.pubkey = "kjfhkefbviury";
        msg.setCandidate(node);
        msg.setLastLogTerm(8956737);
        msg.setLastLogIndex(42956);

        byte[] bytes = RaftMessageParser.setMessage(msg).getBytes();
        RequestVote message = (RequestVote) RaftMessageParser.parse(bytes);

        assert msg.getTerm() == message.getTerm();
        assert msg.getCandidate().pubkey.equals(message.getCandidate().pubkey);
        assert msg.getLastLogTerm() == message.getLastLogTerm();
        assert msg.getLastLogIndex() == message.getLastLogIndex();
    }

    @Test
    public void testRequestVoteResponse() {
        RequestVoteResponse msg = new RequestVoteResponse();
        msg.setTerm(4563657);
        msg.setGranted(true);

        byte[] bytes = RaftMessageParser.setMessage(msg).getBytes();
        RequestVoteResponse message = (RequestVoteResponse) RaftMessageParser.parse(bytes);

        assert msg.getTerm() == message.getTerm();
        assert msg.isGranted() == message.isGranted();
    }

    @Test
    public void testAppendEntries() {
        AppendEntries msg = new AppendEntries();
        msg.setTerm(353);
        PubKeyNode node = new PubKeyNode();
        node.pubkey = "ksuhgkjfjksdbfchj";
        msg.setLeader(node);
        msg.setPrevLogIndex(298476957);
        msg.setPrevLogTerm(5467);
        msg.setLeaderCommit(586793);
        List<LogEntry> entries = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            LogEntry entry = new LogEntry();
            entry.setTerm(i);
            entry.setIndex(i * 100);
            ContractRequest command = new ContractRequest();
            command.setAction(String.format("%20d-%20d", i, i * 100));
            entry.setCommand(command);
            entries.add(entry);
        }
        msg.setEntries(entries);

        byte[] bytes = RaftMessageParser.setMessage(msg).getBytes();
        AppendEntries message = (AppendEntries) RaftMessageParser.parse(bytes);

        assert msg.getTerm() == message.getTerm();
        assert msg.getLeader().pubkey.equals(message.getLeader().pubkey);
        assert msg.getPrevLogIndex() == message.getPrevLogIndex();
        assert msg.getPrevLogTerm() == message.getPrevLogTerm();
        assert msg.getLeaderCommit() == message.getLeaderCommit();
        assert msg.getEntries().size() == message.getEntries().size();
        assert msg.getEntries().get(0).getCommand().getAction().equals(message.getEntries().get(0).getCommand().getAction());
        assert msg.getEntries().get(9).getCommand().getAction().equals(message.getEntries().get(9).getCommand().getAction());
    }

    @Test
    public void testAppendEntriesResponse() {
        AppendEntriesResponse msg = new AppendEntriesResponse();
        msg.setTerm(5349679);
        msg.setSuccess(true);

        byte[] bytes = RaftMessageParser.setMessage(msg).getBytes();
        AppendEntriesResponse message = (AppendEntriesResponse) RaftMessageParser.parse(bytes);

        assert msg.getTerm() == message.getTerm();
        assert msg.isSuccess() == message.isSuccess();
    }
}
